/*
 * BST.c
 * Stephanie Weirich
 * 5/6/94
 */

/***************************************************************************
  This file contains the code for the binary search tree used to contain the
  Z-buffer. Since not all of the Z_buffer needs to  be computed, this stores
  non-intersectign windows of it in a binary tree.
  **************************************************************************/
#include "shade.h"
#include "math.h"
#include <stdio.h>


/***********************
  Tree constructor
  allocates and returns 
  a pointer to a new tree.
  Set's the tree's window 
  to win.
  ************************/
#ifdef ANSI
tree* mkNewTree(Window* win)
#else
     tree* mkNewTree(win)
     Window* win;
#endif
{
  tree* newTree;
  newTree = (tree*) malloc(sizeof(tree));
  newTree->win = win;
  newTree->right = NULL;
  newTree->left = NULL;
  return newTree;
}

/***********************
  Comparative functions 
  **********************/
#ifdef ANSI
bool inWinPoint(Window* old, int x, int y)
#else
bool inWinPoint(old, x, y)
Window* old;
int x, y;
#endif
{
  if ((x >= old->x0) && (x <= old->x1) &&
      (y >= old->y0) && (y <= old->y1))
    return TRUE;
  else
    return FALSE;
}

#ifdef ANSI
bool inWindow(Window* old, Window* new)
#else
     bool inWindow(old, new)
     Window *old, *new;
#endif
{
  if (inWinPoint(old, new->x0, new->y0) || 
      inWinPoint(old, new->x1, new->y0) ||
      inWinPoint(old, new->x0, new->y1) ||
      inWinPoint(old, new->x1, new->y1))
    return TRUE;
  else
    return FALSE;
}

#ifdef ANSI
bool completelyInWindow(Window* old, Window* new)
#else
     bool completelyInWindow(old, new)
     Window *old, *new;
#endif
{
  if (inWinPoint(old, new->x0, new->y0) && 
      inWinPoint(old, new->x1, new->y0) &&
      inWinPoint(old, new->x0, new->y1) &&
      inWinPoint(old, new->x1, new->y1))
    return TRUE;
  else
    return FALSE;
}

/*******************************
  Spilts the window in half to find
  the maximum boundary for each of 
  the children. Alternates between 
  vertical and horizontal splits
  *******************************/


#ifdef ANSI
findChildren(Window* left, Window* right, Window* max, level)
#else
     findChildren(left, right, max, level)
     Window* left, *right, *max;
     int level;
#endif
{
  if (level % 2)
    {
      left->x0 = max->x0;
      left->x1 = (max->x1/2);
      left->y0 = max->y0;
      left->y1 = max->y1;
      right->x0 = (max->x1/2)+1;
      right->x1 = max->x1;
      right->y0 = max->y0;
      right->y1 = max->y1;
    }
  else
    {
      left->x0 = max->x0;
      left->x1 = max->x1;
      left->y0 = max->y0;
      left->y1 = (max->y1/2);
      right->x0 = max->x0;
      right->x1 = max->x1;
      right->y0 = (max->y1/2) +1;
      right->y1 = max->y1; 
    }
}


/**********************************
  Returns the bounding box of the 
  union of win0 and win1 in win0.
  *********************************/
#ifdef ANSI
void stretch(Window* win0, Window* win1)
#else
void stretch(win0, win1)
Window* win1, *win0;
#endif
{
  if (win1->x0 < win0->x0)
    win0->x0 = win1->x0;
  if (win1->x1 > win0->x1)
    win0->x1 = win1->x1;
  if (win1->y0 < win0->y0)
    win0->y0 = win1->y0;
  if (win1->y1 > win0->y1)
    win0->y1 = win1->y1;
}

/*************************************
  Computes the intersection of win0 
  and win0 and returns it in newwin.
  ***********************************/

#ifdef ANSI
void intersection(Window* win0, Window* win1, Window* newwin)
#else
     void intersection (win0, win1, newwin)
     Window *win0, *win1, *newwin;
#endif
{
  if (win1->x0 < win0->x0)
    newwin->x0 = win0->x0;
  else 
    newwin->x0 = win1->x0;
  if (win1->x1 > win0->x1)
    newwin->x1 = win0->x1;
  else 
    newwin->x1 = win1->x1;
  if (win1->y0 < win0->y0)
    newwin->y0 = win0->y0;
  else 
    newwin->y0 = win1->y0;
  if (win1->y1 > win0->y1)
    newwin->y1 = win0->y1;
  else 
    newwin->y1 = win1->y1;
}

/***********************************************
  Adds the window win, to the tree root, who's
  maximum window size is max. Level starts at 0.
  **********************************************/
#ifdef ANSI
void addToTree(tree* root, Window* win, Window* max, int level)
#else
     void addToTree(root, win, max,  level)
     tree* root;
     Window* win, *max;
     int level;
#endif
{
  Window left;
  Window right;
  Window* newleft;
  Window* newright;
  if (root==NULL)
    {
      mkNewTree(win);
    }
  else
    {
      if (inWindow(root->win, win))
	stretch(root->win, win);
      else
	{
	  findChildren (&left, &right, max, level);
	  if (completelyInWindow(&left, win))
	    addToTree(root->left, win, &left, level++);
	  else
	    if (completelyInWindow(&right, win))
	      addToTree(root->right, win, &right, level++);
	    else
	      {
		newleft = (Window*)malloc(sizeof(Window));
		newright = (Window*)malloc(sizeof(Window));
		intersection(left, win, newleft);
		intersection(right, win, newright);
		free(win);
		addToTree(root->left, newleft, &left, level++);
		addToTree(root->right, newright, &right, level++);
	      }
	}
    }
}


/************************************
 Pass it a poly and bounding rect,
 scans it into all appropriate places
 ************************************/
#ifdef ANSI
void ScanTree(tree* root, Window* win, Window* max, Triangle* poly, 
	      Poly_vert ** vlist, int level)
#else
     void ScanTree(root, win, max, poly, vlist, level)
     tree* root;
     Window* win, *max;
     Triangle* poly;
     Poly_vert* vlist;
     int level;
#endif
{
  Window left;
  Window right;
  if (root != NULL)
    {
      if (inWindow(root->win, win))
	poly_scan(vlist, poly, root->win, root->buffer);
      findChildren (&left, &right, max, level);
      if (root->left!=NULL)
	if (inWindow(left, win))
	  ScanTree(root->left, win, left,  poly, vlist, level);
      if (root->right!=NULL)
	if (inWindow(right, win))
	  ScanTree(root->right, win, right, poly, vlist, level);
    }
}

/***********************************
  Returns whether point x,y in with
  the height z, is in shadow in the
  z-buffer.
  ************************************/
#ifdef ANSI
bool InShadow(tree* root, int x, int y, double z, Window* max, level)
#else
     bool InShadow(root, x, y, z, max, level)
     tree* root;
     int x, y;
     Window* max;
     int level;
     double z;
#endif 
{
  Window left;
  Window right;
  bool value1 = FALSE;
  bool value2 = FALSE;
  int my_x, my_y;
  if (root== NULL)
    return FALSE;
  else
    {
      if (inWinPoint(root->win, x, y))
	{
	  my_x = x - root->win->x0;
	  my_y = y - root->win->y0;
	  if (root->buffer[my_x][my_y].y > z)
	    return TRUE;
	  else
	    return FALSE;
	}
      else
	{
	  findChildren (&left, &right, max, level);  
	  if (inWinPoint(right, x, y))
	    value1 = InShadow(root->right, x, y, z, right, level);
	  if (inWinPoint(left, x, y))
	      value2 = InShadow(root->left, x, y, z, left, level);
	  return (value1 || value2);
	}
    }
}

/*************************
  Allocates the Z-buffers 
  after their bounding boxes 
  have been calculated.
  ************************/

#ifdef ANSI
void allocateBuffer(tree* root)
#else
     void allocateBuffer(root)
     tree* root;
#endif
{
  int i,j;
  if (root!=NULL)
    {
      root->xsize = root->win->x1 - root->win->x0 + 1;
      root->ysize = root->win->y1 - root->win->y0 + 1;
      root->buffer = (Y_struct**)malloc(root->xsize*sizeof(Y_struct*));
      if (!root->buffer)
	{
	  printf("Allocate buffers: Out of memory!\n");
	  exit(1);
	}
      for (i =0; i<root->xsize; i++)
	{
	  root->buffer[i] = (Y_struct*)malloc(root->ysize*sizeof(Y_struct));
	  if (!root->buffer[i])
	    {
	      printf("Allocate buffers: Out of memory!\n");
	      exit(1);
	    }
	  for (j=0; j<root->ysize; j++)
	    root->buffer[i][j].y = -1 * HUGE;
	}
      allocateBuffer(root->left);
      allocateBuffer(root->right);
    }
}

/***************************
  Simple tree walk and print
  for debugging purposes.
  **************************/
void PrintBuffers(root)
tree* root;
{
  if (root!= NULL)
    {
      printf("Bounding box: %d %d %d %d\n", root->win->x0, 
	     root->win->y0,root->win->x1,root->win->y1);
      PrintBuffers(root->left);
      PrintBuffers(root->right);
    }
}

int area(root)
tree* root;
{
  int x;
  if (root!=NULL)
    {
      x = (root->win->x1 - root->win->x0) *  (root->win->x1 - root->win->x0);
      x *= area(root->left);
      x *= area(root->right);
      return x;
    }
}


